/*
 * Reads a character via the serial port and
 * sends the "keystroke" to the MK14 via the optical switches 
 *  Uses 9600 baud at present
 *  Note: the input is not case sensitive
 *    It uses 0 to 9 and a to f for the standard keys  
 *    m for mem key
 *    t for the term key
 *    q for the abort key ( as A already ysed )
 *    r to reset the MK14
 *    
 *    send x to turn on the debug stuff whcih sends data back via the serial port
 *    
 * If using the arduino IDE  Serial monitor you can send a string of characters 
 * e.g. rm0f20tc4tmtfet 
 *  will do a reset and set memory location F20 to C4 and F21 to FE
 *  Note: This is using the older version of the MK14 
 *  
 *  
 *  copyright David Allday April 2020 
 *  
  */

/*
 * Edge connector starting at reset button end
 * 0v
 * 0v
 * col 6 (6 ,E )
 * col 7 (7, F, Term )
 * col 5 (5)
 * not used
 * col 4 (4, Abort )
 * not used
 * col 3 (3, D, Mem )
 * col 2 (2, C, Go )
 * col 1 (1, 9, B )
 * col 0 (0, 8, A )
 * row 3 (Go, Mem, Abort, Term )
 * row 2 (A, B, C, D, E, F )
 * row 1 (8, 9)
 * row 0 (0, 1, 2, 3, 4, 5, 6, 7 )
 * 
 * To activate a key then the correct row and column need to be connected.
 * This is done by taking the relevant row pin low and column pin low.
 * 
 * The reset is done by shorting out the reset capacitor and that is achieve by take the reset pin low.
 * 
 */

// define which pins are used for columns and rows
//  

// these are for the Arduino Uno 
#define row0 A0
#define row1 A1
#define row2 A2
#define row3 A3

#define col0 2
#define col1 3
#define col2 4
#define col3 5
#define col4 6
#define col5 7
#define col6 9  // columns 6 and 7 seem to be switched on the edge connector
#define col7 8

// pin used for the reset
#define ResetPin A5 //Changed from pin 10 originally


// these are for the Mega256
/*
#define row0 29
#define row1 31
#define row2 33
#define row3 35

#define col0 37
#define col1 39
#define col2 41
#define col3 43
#define col4 45
#define col5 47
#define col6 53  // columns 6 and 7 seem to be switched on the edge connector
#define col7 49

// pin used for the reset
#define ResetPin 27
*/

#define PinWaitingForInput 13

// Minimum length of one keypress (seconds)
#define KeyPressLength 10       // 10  == 0.01 second 
// Time for post-release debounce delay (seconds).
#define KeyReleaseLength 5   // 5  == 0.005 second 

bool debugMode = false;

void setup() {
  // Use the serial port 
  Serial.begin(9600);


  // first set the output values to high 
  // so that once switched to output they start high
  digitalWrite(row0, HIGH);
  digitalWrite(row1, HIGH);
  digitalWrite(row2, HIGH);
  digitalWrite(row3, HIGH);

  digitalWrite(col0, HIGH);
  digitalWrite(col1, HIGH);
  digitalWrite(col2, HIGH);
  digitalWrite(col3, HIGH);
  digitalWrite(col4, HIGH);
  digitalWrite(col5, HIGH);
  digitalWrite(col6, HIGH);
  digitalWrite(col7, HIGH);
  
  digitalWrite(10,HIGH); //Disable SRAM for now.
  digitalWrite(ResetPin, HIGH);

  // Serial.println("all pins set to high");
  // set the row and column and reset pins as outputs
  pinMode(row0,OUTPUT);
  pinMode(row1,OUTPUT);
  pinMode(row2,OUTPUT);
  pinMode(row3,OUTPUT);

  pinMode(col0,OUTPUT);
  pinMode(col1,OUTPUT);
  pinMode(col2,OUTPUT);
  pinMode(col3,OUTPUT);
  pinMode(col4,OUTPUT);
  pinMode(col5,OUTPUT);
  pinMode(col6,OUTPUT);
  pinMode(col7,OUTPUT);
  
  pinMode(ResetPin,OUTPUT);

  // Serial.println("all pins set to output");

}

void loop() {
    // main code
    char keyToPress='0';
    bool displayKey=true;  // fix to display character if not cr etc.
    digitalWrite(PinWaitingForInput, HIGH);
    // Serial.print (">");
    keyToPress = ReadSerialCharacter();
    digitalWrite(PinWaitingForInput, LOW);
    
    switch ( keyToPress )
    {
      case 'M':        // Mem, the most commonly struck key
      case 'm':
          PressKey(row3,col3);
          break;
      case '0':
          PressKey(row0,col0);
          break;
      case '1':
          PressKey(row0,col1);
          break;
      case '2':
          PressKey(row0,col2);
          break;
      case '3':
          PressKey(row0,col3);
          break;
      case '4':
          PressKey(row0,col4);
          break;
      case '5':
          PressKey(row0,col5);
          break;
      case '6':
          PressKey(row0,col6);
          break;
      case '7':
          PressKey(row0,col7);
          break;
      case '8':
          PressKey(row1,col0);
          break;
      case '9':
          PressKey(row1,col1);
          break;
      case 'A':
      case 'a':
          PressKey(row2,col0);
          break;
      case 'B':
      case 'b':
          PressKey(row2,col1);
          break;
      case 'C':
      case 'c':
          PressKey(row2,col2);
          break;
      case 'D':
      case 'd':
          PressKey(row2,col3);
          break;
      case 'E':
      case 'e':
          PressKey(row2,col6);
          break;
      case 'F':
      case 'f':
          PressKey(row2,col7);
          break;
      case 'G':      // Go
      case 'g':
          PressKey(row3,col2);
          break;
      case 'Q':     // Q=Abort ('A' already taken)
      case 'q':
          PressKey(row3,col4);
          break;
      case 'T':     // Term
      case 't':
          PressKey(row3,col7);
          break;

      case 'R':     // System Reset
      case 'r':
          if ( debugMode){
              Serial.print(" reset pin pulsed ");
          }
          digitalWrite(ResetPin, LOW);
          delay(100);    // delay .1 second = 100 milliseconds 
          digitalWrite(ResetPin, HIGH);
          delay(1000); // give it time to actually reset
          break;
      case 'x':    // turn on debug mode
          displayKey=false;
          debugMode = true;
          Serial.println("");
          Serial.println("Debug mode on");
          break;
      // new line
      case 10:   // line feed
        displayKey=false;
        if ( ! debugMode){
            Serial.println("");
        }
          break;
      // ignore a some noise characters    
      case 12:   // carrage return
          displayKey=false;
          break;
      case ' ':  // space
          break;

      default: 
          displayKey=false;
          if (debugMode){
              Serial.print (" unknown key");
          }
          break; 

    }
    if (displayKey){
        Serial.print(char(keyToPress));
    }
    
    if (debugMode){
      Serial.println("");
    }
}

// Momentarily connect the specified key row / key column
// together by activating their associated GPIO port pins
void PressKey(int RowPin, int ColumnPin){

    if (debugMode){
      Serial.print(" ");
      Serial.print("Row Pin ");
      Serial.print(RowPin);
      Serial.print(" Column Pin ");
      Serial.print(ColumnPin);
      Serial.print(" ");
    }
    digitalWrite(RowPin,LOW);
    digitalWrite(ColumnPin,LOW);
    delay(KeyPressLength);
    digitalWrite(ColumnPin,HIGH);
    digitalWrite(RowPin,HIGH);
    delay(KeyReleaseLength);
}


char ReadSerialCharacter()
{
  byte c;
  do 
  {
    if (Serial.available())
    {
      c = Serial.read();
      return c;
    }
  } while (true);
}
